-- the package contains types, constants, and function prototypes
package cpu_package is

-- ALTER THE FOLLOWING CONSTANTS AND TYPE DEFINITIONS 
-- TO SUIT YOUR PROCESSOR

-- define the physical sizes in terms of bits 
  constant word_size: positive:=16; -- width of data bus
  constant address_size: positive:=12; -- width of address bus
  type address_val is range 0 to (2**address_size)-1; -- valid address range

-- define the ram and rom sizes and locations
  constant ram_size: positive:=2**8; -- size of ram
  constant rom_size: positive:=2**8; -- size of rom
  constant rom_bottom: address_val := 0; -- start of rom
  constant rom_top: address_val := address_val(rom_size-1);
  constant ram_bottom: address_val := address_val(2**address_size-ram_size); 
                                      -- ram is at top of address map
  constant ram_top: address_val := address_val(2**address_size-1);

-- define subparts of an instruction
  subtype opcode is bit_vector(3 downto 0);
  subtype source is bit_vector(3 downto 0);
  subtype target is bit_vector(3 downto 0);
  subtype dest is bit_vector(3 downto 0);
  subtype imm_value is bit_vector(7 downto 0);
  subtype offset is bit_vector(7 downto 0);
  subtype mode is bit_vector(3 downto 0);
   
-- define the valid opcodes  
  constant op_nop:   opcode:= B"0000";      
  constant op_add:   opcode:= B"0001";
  constant op_sub:   opcode:= B"0010";
  constant op_store: opcode:= B"0011";
  constant op_load:  opcode:= B"0100";
  constant op_branch: opcode:= B"0101";

-- define the addressing modes
  constant mode_imm:   mode:= B"0001";
  constant mode_ind:   mode:= B"0010";
  constant mode_abs:   mode:= B"0011";

-- define the number of general registers
  subtype word is bit_vector(word_size-1 downto 0);
  type register_file is array(0 to 5) of word;

-- give names to the registers (optional)
  constant r0: natural := 0;
  constant r1: natural := 1;
  constant r2: natural := 2;
  constant r3: natural := 3;
  constant r4: natural := 4;
  constant r5: natural := 5;
  
-- YOU ARE ADVISED NOT TO CHANGE ANYTHING BELOW HERE
-- UNLESS YOU ARE VERY CONFIDENT ABOUT IT

  subtype address is bit_vector(address_size-1 downto 0);
  type program_array is array(0 to rom_size-1) of word;
  type ram_array is array(0 to ram_size-1) of word;

-- resolution function for data bus
  type word_array is array (integer range ) of word;
  function resolve_word(driver: word_array)return word; 
  subtype bus_word is resolve_word word;

-- define the numeric range of the different types
  type signed is range -(2**(word_size-1)) to 2**(word_size-1)-1;
  type unsigned is range 0 to (2**word_size)-1;



-- utility functions to translate between bit vectors and values
  function vec_to_nat(bv: bit_vector)return natural;
  function vec_to_int(bv: bit_vector)return integer;
  function nat_to_address(address_var: address_val)return address;
  function address_to_nat(bit_address: address)return address_val;
  function word_to_int(bit_data: word)return integer;
  function int_to_word(data_var: integer)return word;

end cpu_package;

-- the package body contains the actual functions
package body cpu_package is
  
  --convert a bit_vector of any length to a natural
  
  function vec_to_nat(bv: bit_vector)return natural is
      variable val: natural := 0;
      begin
         for i in bv'range loop
            if bv(i) = '1' then
                val := val + 2**i;
            end if; 
         end loop;
      return val;
  end vec_to_nat;
  
  --convert a bit_vector of any length to an integer 
  
  function vec_to_int(bv: bit_vector)return integer is
      variable val: integer := 0;
      variable negative: boolean;
      begin
         -- if MSB = 1 must be negative
         negative := bv(bv'left)='1';
         for i in bv'range loop
            if bv(i) = '1' then
                val := val + 2**i;
            end if; 
         end loop;
         if negative then -- 2's complement result
            val := (-val)-1;
         end if;
      return val;
  end vec_to_int;

-- convert a numeric address to a bit_vector
  function nat_to_address(address_var: address_val)return address is
      variable address_vector: address;
      variable tmp: address_val;
      begin
         tmp := address_var;
         for i in 0 to (address_size-1) loop
             if (tmp mod 2) = 1 then
                address_vector(i) := '1';
             else address_vector(i) := '0';
             end if;
             tmp := tmp / 2;
         end loop;
     return address_vector;
  end nat_to_address;

  --convert a bit_vector to an address
  function address_to_nat(bit_address: address)return address_val is
      variable address_value: address_val := 0;
      begin
         for i in address'range loop
            if bit_address(i) = '1' then
                address_value := address_value + 2**i;
            end if; 
         end loop;
      return address_value;
  end address_to_nat;
     
  --convert a data bit_vector to an integer 
  function word_to_int(bit_data: word)return integer is
      variable data_value: integer:= 0;
      variable negative: boolean;
      begin
         -- if MSB = 1 must be negative
         negative := bit_data(bit_data'left)='1';
         for i in word'range loop
            data_value := data_value * 2;
            if bit_data(i) = '1' then
                data_value := data_value + 1;
            end if; 
         end loop;
         if negative then -- 2's complement result
            data_value := (-data_value)-1;
         end if;
      return data_value;
  end word_to_int;

  -- convert an integer to a data bit_vector
  function int_to_word(data_var: integer)return word is
      variable data_vector: word;
      variable tmp: integer;
      variable negative: boolean;
      begin
         -- 2's complement if input negative
         negative :=  data_var < 0;
         if negative then   -- 2's complement number
            tmp := -(data_var+1);
         else tmp := data_var;
         end if;
         -- convert 2's complement number to binary
         for i in 0 to (word_size-1) loop
             if (tmp mod 2) = 1 then
                data_vector(i) := '1';
             else data_vector(i) := '0';
             end if;
             tmp := tmp / 2;
         end loop;
     return data_vector;
  end int_to_word; 

  -- simple resolution function for the data bus
  -- this just 'or's all signals on the bus
  -- no tristate logic is allowed for
  function resolve_word(driver: word_array)return word is
      variable result: word;
  begin
      -- set up default value of 0
      for i in word'range loop
          result(i) := '0';
      end loop;
      -- resolve the inputs
      for i in driver'range loop
          result := result or driver(i);
      end loop;
  return result;
  end resolve_word;

end cpu_package; 